#include "procmap.h"

#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGE, PmUnload)
#pragma alloc_text(PAGE, PmCreateNotifyRoutine)
#pragma alloc_text(PAGE, PmMapProcessName)
#pragma alloc_text(PAGE, PmGetImageNameOffset)
#pragma alloc_text(PAGE, Pm_MJ_Create)
#pragma alloc_text(PAGE, Pm_MJ_Close)
#pragma alloc_text(PAGE, Pm_MJ_DeviceControl)

NTSTATUS DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath
) {
    NTSTATUS Status;
    UNICODE_STRING NtName, DosName;
    PDEVICE_OBJECT DeviceObject = NULL;
    UNREFERENCED_PARAMETER(RegistryPath);
    
    #if DBGPRINT
    DbgPrint("PmDriverEntry\n");
    #endif

    if (!PmGetImageNameOffset())
        return STATUS_INTERNAL_ERROR;
        
    RtlInitUnicodeString(&NtName, PROCMAP_NT_DEVICE_NAME);
    
    // Create the device    
    Status = IoCreateDevice(
                            DriverObject,
                            0,
                            &NtName,
                            FILE_DEVICE_UNKNOWN,
                            0,
                            0,
                            &DeviceObject
                           );
                           
    if (Status != STATUS_SUCCESS)
    {
        #if DBGPRINT
        DbgPrint("PmDriverEntry: IoCreateDevice status = 0x%08x\n", Status);
        #endif
        return Status;
    }
    
    RtlInitUnicodeString(&DosName, PROCMAP_DOS_DEVICE_NAME);
    
    Status = IoCreateSymbolicLink(&DosName, &NtName);
    
    if (Status != STATUS_SUCCESS)
    {
        // Cleanup the device
        IoDeleteDevice(DriverObject->DeviceObject);
        #if DBGPRINT
        DbgPrint("PmDriverEntry: IoCreateSymbolicLink status = 0x%08x\n", Status);
        #endif
        return Status;
    }
    
    // Register device function callbacks
    DriverObject->MajorFunction[IRP_MJ_CREATE] = Pm_MJ_Create;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = Pm_MJ_Close;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = Pm_MJ_DeviceControl;
        
    DriverObject->DriverUnload = PmUnload;
    
    // Registrer process create callback
    PsSetCreateProcessNotifyRoutine(PmCreateNotifyRoutine, 0);

    return Status;
}

void PmUnload(
    IN PDRIVER_OBJECT DriverObject
) {
    UNICODE_STRING DosName;
    
    PAGED_CODE();

    #if DBGPRINT
    DbgPrint("PmUnload\n");
    #endif

    PsSetCreateProcessNotifyRoutine(PmCreateNotifyRoutine, 1);

    // Cleanup the device
    RtlInitUnicodeString(&DosName, PROCMAP_DOS_DEVICE_NAME);
    IoDeleteSymbolicLink(&DosName);
    IoDeleteDevice(DriverObject->DeviceObject);
    
    return;
}

void PmCreateNotifyRoutine(
    IN HANDLE  ParentId,
    IN HANDLE  ProcessId,
    IN BOOLEAN  Create
) {
    PAGED_CODE();

    if (Create)
    {
        HANDLE ProcessHandle = NULL;
        CLIENT_ID ClientId = {0};
        OBJECT_ATTRIBUTES ObjectAttributes;

        #if DBGPRINT
        DbgPrint("PmCreateNotifyRoutine: ProcessId = %u\n", ProcessId);
        #endif
        ClientId.UniqueProcess = ProcessId;

        InitializeObjectAttributes(
                                    &ObjectAttributes,
                                    NULL,
                                    0,
                                    NULL,
                                    NULL
                                  );
        ZwOpenProcess(
                        &ProcessHandle,
                        PROCESS_ALL_ACCESS,
                        &ObjectAttributes,
                        &ClientId
                     );

        if (ProcessHandle)
        {
            PEPROCESS Process = NULL;

            #if DBGPRINT
            DbgPrint("PmCreateNotifyRoutine: ProcessHandle = %u\n", ProcessHandle);
            #endif

            ObReferenceObjectByHandle(
                                        ProcessHandle,
                                        PROCESS_ALL_ACCESS,
                                        NULL,
                                        KernelMode,
                                        &Process,
                                        NULL
                                     );
            if (Process)
            {
                char *ImageName = ((char *) Process) + PmGetImageNameOffset();

                #if DBGPRINT
                DbgPrint("PmCreateNotifyRoutine: Process = 0x%08x, ImageName = %s\n", Process, ImageName);
                #endif
                PmMapProcessName(ImageName);
                ObDereferenceObject(Process);
            }

            ZwClose(ProcessHandle);
        }
    }
}

void PmMapProcessName(
    char *ImageName
) {
    OBJECT_ATTRIBUTES ObjectAttributes;
    UNICODE_STRING KeyName;
    HANDLE Key = NULL;

    PAGED_CODE();
    
    // Open the registry keys
    RtlInitUnicodeString(&KeyName, L"\\REGISTRY\\Machine\\SOFTWARE\\ProcMap");

    InitializeObjectAttributes(
                                &ObjectAttributes,
                                &KeyName,
                                OBJ_CASE_INSENSITIVE,
                                NULL,
                                NULL
                              );

    ZwOpenKey(
                &Key,
                KEY_READ,
                &ObjectAttributes
             );

    // Look for the value
    if (Key)
    {
        ULONG Index = 0,
              Length;
        NTSTATUS Status = 0;
        char KeyBuf[(sizeof(KEY_VALUE_FULL_INFORMATION) - sizeof(WCHAR)) + ((MAX_IMAGE_NAME_LEN + MAX_IMAGE_NAME_LEN + 1) * sizeof(WCHAR))];

        // Check all registry values
        do
        {
            RtlFillMemory(&KeyBuf[0], sizeof(KeyBuf), 0);
            Length = sizeof(KeyBuf);
            Status = ZwEnumerateValueKey(
                                          Key,
                                          Index,
                                          KeyValueFullInformation,
                                          &KeyBuf[0],
                                          Length,
                                          &Length
                                        );

            #if DBGPRINT
            DbgPrint(
                        "PmMapProcessName: ZwEnumerateValueKey Status 0x%08x, NameLength %u, DataLength %u\n",
                        Status,
                        ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->NameLength,
                        ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->DataLength
                    );
            #endif

            if (!Status &&
                ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->Type == REG_SZ &&
                ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->NameLength <= (MAX_IMAGE_NAME_LEN << 1) &&
                ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->DataLength <= ((MAX_IMAGE_NAME_LEN + 1) * sizeof(WCHAR)))
            {
                ANSI_STRING ImageString,
                            RegString;

                UNICODE_STRING ValueString;

                char RegBuf[MAX_IMAGE_NAME_LEN + 1];

                ValueString.Buffer = &((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->Name[0];
                ValueString.Length = (USHORT) ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->NameLength;
                ValueString.MaximumLength = ValueString.Length;

                // Setup a string for converting the registry value name to ansi
                RtlFillMemory(&RegBuf[0], sizeof(RegBuf), 0);
                RegString.Length = 0;
                RegString.MaximumLength = sizeof(RegBuf);
                RegString.Buffer = &RegBuf[0];

                RtlUnicodeStringToAnsiString(
                                                &RegString,
                                                &ValueString,
                                                0
                                            );

                #if DBGPRINT
                DbgPrint("PmMapProcessName: RegString.Buffer %s\n", RegString.Buffer);
                DbgPrint("PmMapProcessName: RegString.Length %u\n", RegString.Length);
                #endif

                ImageString.Buffer = ImageName;
                // Get the length of the image name
                ImageString.Length = 0;

                while (ImageString.Buffer[ImageString.Length])
                    ImageString.Length++;

                ImageString.MaximumLength = sizeof(RegBuf);

                Status = RtlCompareString(
                                            &RegString,
                                            &ImageString,
                                            1
                                         );

                if (!Status)
                {
                    // Get the value and fix the image name
                    ValueString.Buffer = (WCHAR *) ((ULONG) &KeyBuf[0] + ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->DataOffset);
                    ValueString.Length = (USHORT) ((PKEY_VALUE_FULL_INFORMATION) &KeyBuf[0])->DataLength;
                    ValueString.MaximumLength = ValueString.Length;

                    // Setup a string for converting the registry value name to ansi
                    RegString.Length = 0;
                    RegString.MaximumLength = sizeof(RegBuf);
                    RegString.Buffer = ImageName;
                    RtlFillMemory(RegString.Buffer, sizeof(RegBuf), 0);

                    RtlUnicodeStringToAnsiString(
                                                    &RegString,
                                                    &ValueString,
                                                    0
                                                );
                    #if DBGPRINT
                    DbgPrint("PmMapProcessName: ImageName -> %s\n", RegString.Buffer);
                    #endif
                    break;
                }
                else
                {
                    Status = STATUS_SUCCESS;
                }
            }

            Index++;
        }
        while (!Status || Status == STATUS_BUFFER_TOO_SMALL);

        ZwClose(Key);
    }
}

ULONG PmGetImageNameOffset(
    void
) {
    static ULONG Offset = 0;
    HANDLE ProcessHandle = NULL;
    PEPROCESS Process = NULL;
    CLIENT_ID ClientId = {0};
    OBJECT_ATTRIBUTES ObjectAttributes;
    ULONG Index = 0;
    
    PAGED_CODE();
    
    // Get the ImageFileName from the EPROCESS struct by looking
    // for the name of the "System" process
    if (Offset == 0)
    {
        do
        {
            ClientId.UniqueProcess = (HANDLE) ++Index;
            ClientId.UniqueThread = (HANDLE) 0;
        
            InitializeObjectAttributes(
                                        &ObjectAttributes,
                                        NULL,
                                        0,
                                        NULL,
                                        NULL
                                      );
            ProcessHandle = NULL;
            ZwOpenProcess(
                            &ProcessHandle,
                            PROCESS_ALL_ACCESS,
                            &ObjectAttributes,
                            &ClientId
                         );
    
       
            if (ProcessHandle)
            {
                Process = NULL;
        
                ObReferenceObjectByHandle(
                                            ProcessHandle,
                                            PROCESS_ALL_ACCESS,
                                            NULL,
                                            KernelMode,
                                            &Process,
                                            NULL
                                         );

                if (Process)
                {
                    char *Temp = (char *) Process;
                    char SystemName[] = "System";
                    
                    do 
                    {
                        if (RtlCompareMemory(SystemName, Temp, sizeof(SystemName) - 1) != (sizeof(SystemName) - 1))
                            Temp += 4;
                        else
                            Offset = (ULONG) Temp - (ULONG) Process;
                    }
                    while (Offset == 0);
                    
                    ObDereferenceObject(Process);
                }
        
                ZwClose(ProcessHandle);
                
                #if DBGPRINT
                DbgPrint("PmGetImageNameOffset: System PID = %u, Offset = 0x%x\n", ClientId.UniqueProcess, Offset);
                #endif
    
            }
        }
        while (Offset == 0 && Index < 0x000011);
    }
   
    return Offset;
}

NTSTATUS Pm_MJ_Create(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
) {
    PAGED_CODE();
    
    #if DBGPRINT
    DbgPrint("Pm_MJ_Create: IOCTL_PROCMAP_TERMINATE_PROCESS = 0x%08x\n", IOCTL_PROCMAP_TERMINATE_IMAGE);
    #endif
    
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    
    return STATUS_SUCCESS;
}

NTSTATUS Pm_MJ_Close(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
) {
    PAGED_CODE();
    
    #if DBGPRINT
    DbgPrint("Pm_MJ_Close\n");
    #endif
    
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    
    return STATUS_SUCCESS;
}

NTSTATUS Pm_MJ_DeviceControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
) {
    NTSTATUS RetStatus = STATUS_UNSUCCESSFUL;
    PIO_STACK_LOCATION IoStack;
    
    PAGED_CODE();
    
    IoStack = IoGetCurrentIrpStackLocation(Irp);
    Irp->IoStatus.Information = 0;
    
    switch (IoStack->Parameters.DeviceIoControl.IoControlCode)
    {
        case IOCTL_PROCMAP_TERMINATE_PROCESS:
        {
            HANDLE ProcessHandle = NULL;
            CLIENT_ID ClientId = {0};
            OBJECT_ATTRIBUTES ObjectAttributes = {0};
    
            if (IoStack->Parameters.DeviceIoControl.InputBufferLength != sizeof(HANDLE))
            {
                RetStatus = STATUS_INVALID_PARAMETER;
                break;
            }
            
            ClientId.UniqueProcess = *(HANDLE *) Irp->AssociatedIrp.SystemBuffer;
            
            #if DBGPRINT
            DbgPrint("Pm_MJ_DevcieControl: IOCTL_PROCMAP_TERMINATE_PROCESS %u\n", ClientId.UniqueProcess);
            #endif
    
            InitializeObjectAttributes(
                                        &ObjectAttributes,
                                        NULL,
                                        0,
                                        NULL,
                                        NULL
                                      );
            RetStatus = ZwOpenProcess(
                                        &ProcessHandle,
                                        0x00000001, // PROCESS_TERMINATE
                                        &ObjectAttributes,
                                        &ClientId
                                     );
                         
            if (RetStatus == STATUS_SUCCESS)
            {
                RetStatus = ZwTerminateProcess(ProcessHandle, 1);
                if (RetStatus == STATUS_SUCCESS)
                {
                    #if DBGPRINT
                    DbgPrint("Pm_MJ_DevcieControl: IOCTL_PROCMAP_TERMINATE_PROCESS PID = %u, Status = 0x%08x\n", ClientId.UniqueProcess, RetStatus);
                    #endif
                }
            }
            else if (RetStatus == STATUS_INVALID_PARAMETER)
            {
                RetStatus = STATUS_NO_SUCH_FILE;
            }
            
            if (ProcessHandle)
                ZwClose(ProcessHandle);
            break;
        }
        
        case IOCTL_PROCMAP_TERMINATE_IMAGE:
        {
            NTSTATUS Status = STATUS_SUCCESS;
            HANDLE ProcessHandle = NULL;
            PEPROCESS Process = NULL;
            CLIENT_ID ClientId = {0};
            OBJECT_ATTRIBUTES ObjectAttributes = {0};
            ULONG Index = 1, Count = 0;
            ANSI_STRING ImageString = {0}, TargetString = {0};
            char TargetName[MAX_IMAGE_NAME_LEN + 1] = {0};
            long Compare = 0;
    
            if (IoStack->Parameters.DeviceIoControl.InputBufferLength < 2)
            {
                RetStatus = STATUS_INVALID_PARAMETER;
                break;
            }
            
            RtlCopyMemory(
                            TargetName,
                            Irp->AssociatedIrp.SystemBuffer,
                            (
                                IoStack->Parameters.DeviceIoControl.InputBufferLength < MAX_IMAGE_NAME_LEN 
                                ?
                                IoStack->Parameters.DeviceIoControl.InputBufferLength
                                :
                                MAX_IMAGE_NAME_LEN
                            )
                         );
                         
            TargetString.Buffer = TargetName;
            // Get the length of the image name
            TargetString.Length = 0;

            while (TargetString.Buffer[TargetString.Length])
                TargetString.Length++;

            TargetString.MaximumLength = sizeof(TargetName);
            
            #if DBGPRINT
            DbgPrint("Pm_MJ_DevcieControl: IOCTL_PROCMAP_TERMINATE_IMAGE %s\n", TargetName);
            #endif
          
            do
            {
                InitializeObjectAttributes(
                                            &ObjectAttributes,
                                            NULL,
                                            0,
                                            NULL,
                                            NULL
                                          );
                
                ClientId.UniqueProcess = (HANDLE) Index;
                ClientId.UniqueThread = (HANDLE) 0;
                
                RetStatus = ZwOpenProcess(
                                            &ProcessHandle,
                                            0x00000001, // PROCESS_TERMINATE
                                            &ObjectAttributes,
                                            &ClientId
                                         );
                             
                if (RetStatus == STATUS_SUCCESS)
                {
                    RetStatus = ObReferenceObjectByHandle(
                                                            ProcessHandle,
                                                            PROCESS_ALL_ACCESS,
                                                            NULL,
                                                            KernelMode,
                                                            &Process,
                                                            NULL
                                                          );
                    if (RetStatus == STATUS_SUCCESS)
                    {
                        char *ImageName = ((char *) Process) + PmGetImageNameOffset();

                        // Compare the name
                        ImageString.Buffer = ImageName;
                        // Get the length of the image name
                        ImageString.Length = 0;
            
                        while (ImageString.Buffer[ImageString.Length])
                            ImageString.Length++;
            
                        ImageString.MaximumLength = sizeof(TargetName);
                        
                        Compare = RtlCompareString(
                                                    &TargetString,
                                                    &ImageString,
                                                    1
                                                  );

                        ObDereferenceObject(Process);
                                                  
                        if (Compare == 0)
                        {
                            Status = ZwTerminateProcess(ProcessHandle, 1);
                            if (Status == STATUS_SUCCESS)
                            {
                                Count++;
                                #if DBGPRINT
                                DbgPrint("Pm_MJ_DevcieControl: IOCTL_PROCMAP_TERMINATE_IMAGE Index = %u, ImageName = %s, Count = %u, Status = 0x%08x\n", Index, ImageName, Count, Status);
                                #endif
                            }
                        }
                    }
                }
                
                if (ProcessHandle)
                    ZwClose(ProcessHandle);
                    
                Index++;
            }
            while (Index < 0x10000);
            
            if (Count)
                RetStatus = STATUS_SUCCESS;
            else if (!Count && Status != STATUS_SUCCESS)
                RetStatus = Status;
            else
                RetStatus = STATUS_NO_SUCH_FILE;
            
            break;
        }        
            
        default:
            #if DBGPRINT
            DbgPrint("Pm_MJ_DevcieControl: IOCTL_PROCMAP_UNKNOWN\n");
            #endif
            RetStatus = STATUS_INVALID_PARAMETER;
            break;
    }
    
    Irp->IoStatus.Status = RetStatus;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return RetStatus;
}
